
/** 
 * @defgroup Insert CUDA Memory Advice
 *
 * @brief Set and unset CUDA memory-advice hints about the usage of managed-memory buffers involved in cudarizable with loops. 
 * 
 * High-level explanation:
 *  In case CUDA unified memory is used (-t cuda_man or -t multi_gpu), this sub-phase is run in the memory phase to insert CUDA
 *  memory-advice hints before and after cudarizable with-loops. In general, a CUDA programmer can set and unset memory hints
 *  to CUDA unified memory buffers, which act as heuristics for the CUDA driver. Currently, this sub-phase only tries to insert
 *  ReadMostly memory advice in a naive manner. Simply, we look at a cudarizable with-loop, and if we detect that it reads from
 *  some input array, then this input array will be set to "ReadMostly" just before the with-loop, and this hint will be unset
 *  right after. Note that "ReadMostly" is CUDA terminology, because technically you can still write to these buffers, but since
 *  write bandwidth will be sacrificed for significantly increased read bandwidth, this should be avoided at all cost. Arrays read
 *  from in with-loops are ofcourse read-only for the duration of the with-loop, but to keep it consistent with CUDA terminology
 *  we still call it "ReadMostly". 
 * 
 * Implementation:
 *  Briefly, to do what is mentioned in the previous paragraph, we traverse cudarizable with-loops and look for "idx_sel"
 *  primitives, which are used to select/read/index from arrays. When found, we check the second argument, which is the array id
 *  and verify that we have not encountered an "idx_sel" primitive on this array yet. If so, we generate "set_read_mostly" and
 *  "unset_read_mostly" primitives for that array, and add it to pre- and postassign chains that are at the end of the with-loop
 *  traversal added just before and right after the with-loop.
 * 
 *  The ICM's for the new "set_read_mostly" and "unset_read_mostly" primitives are generated in codegen/compile.c, in "COMPprfSetReadMostly"
 *  and "COMPprfUnsetReadMostly", respectively. These ICM's are compiled in codegen/icm2c_cuda.c to a C ICM, which is defined in
 *  runtime/cuda_h/cuda.h (SAC_CUDA_MEM_ADVISE(...)).
 * 
 * For example:
 *  
 *  a_mem = ...;
 *  b_mem = with {
 *          ([l0,l1,2] <= iv = [eat1, eat2, eat3](IDXS:_wlidx) < [u0,u1,u2])
 *          #pragma gpukernel ...
 *          {
 *             _iv1 = _fill_(_idxs2offset(shp, eat1, eat2, eat3));
 *             read_res = _fill_(_idx_sel(_iv1, a_mem));                                     // <- idx_sel into a_mem
 *             _free_(iv1);
 *             final_res = ...;
 *             _free_(read_res);            
 *          } : final_res;
 *        } : genarray(shp, a_mem);
 * 
 *  ==>
 * 
 *  a_mem = ...;
 *  _set_read_mostly_(a_mem);                                                                // <- set read mostly on a_mem
 *  b_mem = with {
 *          ([l0,l1,2] <= iv = [eat1, eat2, eat3](IDXS:_wlidx) < [u0,u1,u2])
 *          #pragma gpukernel ...
 *          {
 *             _iv1 = _fill_(_idxs2offset(shp, eat1, eat2, eat3));
 *             read_res = _fill_(_idx_sel(_iv1, a_mem));
 *             _free_(iv1);
 *             final_res = ...;
 *             _free_(read_res);            
 *          } : final_res;
 *        } : genarray(shp, b_mem);
 *  _unset_read_mostly_(a_mem);                                                              // <- unset read mostly on a_mem
 * 
 * @ingroup mem
 * 
 * @{
 */


/** <!--********************************************************************-->
 *
 * @file insert_memory_advice.c
 *
 * Prefix: ICMA
 * 
 * debug with (-doICMA optional depending on whether the optimisation is on by default or not): 
 * <SAC2C> -t cuda_man  <src> -#d,ICMA -doICMA
 * <SAC2C> -t multi_gpu <src> -#d,ICMA -doICMA
 *****************************************************************************/
#include "insert_memory_advice.h"
#include "tree_basic.h"
#include "tree_compound.h"
#include "traverse.h"
#include "memory.h"
#include "free.h"
#include "ctinfo.h"
#include "print.h"
#include "globals.h"
#include "cuda_utils.h"
#include "new_types.h"
#include "LookUpTable.h"
#include "DupTree.h"
#include "str.h"

#define DBUG_PREFIX "ICMA"

#include "debug.h"

/** <!--********************************************************************-->
 *
 * @name INFO structure
 * @{
 *
 *****************************************************************************/
struct INFO {
    // flag that signals we are currently traversing a cuda with-loop
    bool in_cudawl;
    // flag that signals we are currently traversing an array-selection primitive
    bool in_prfsel;
    /**
     * look-up table that is used to keep track of which array selections have
     * already been encountered before to prevent duplicate memory advice
     */  
    lut_t *lut;
    // these assigns set memory advice for current with-loop
    node *preassigns;
    // these assigns unset memory advice for current with-loop
    node *postassigns;
};

#define INFO_INCUDAWL(n) (n->in_cudawl)
#define INFO_INPRFSEL(n) (n->in_prfsel)
#define INFO_LUT(n) (n->lut)
#define INFO_PREASSIGNS(n) (n->preassigns)
#define INFO_POSTASSIGNS(n) (n->postassigns)

/** 
 * @name INFO functions
 *
 * @{
 */
static info *
MakeInfo (void)
{
    info *result;

    DBUG_ENTER ();

    result = (info *)MEMmalloc (sizeof (info));

    INFO_INCUDAWL (result) = FALSE;
    INFO_LUT (result) = NULL;
    INFO_INPRFSEL(result) = FALSE;
    INFO_PREASSIGNS(result) = NULL;
    INFO_POSTASSIGNS(result) = NULL;

    DBUG_RETURN (result);
}

static info *
FreeInfo (info *info)
{
    DBUG_ENTER ();

    info = MEMfree (info);

    DBUG_RETURN (info);
}
/** <!--********************************************************************-->
 * @}  <!-- INFO structure -->
 *****************************************************************************/

/** <!--********************************************************************-->
 *
 * @name Static helper functions
 * @{
 *
 *****************************************************************************/

/** <!--********************************************************************-->
 *
 * @fn static void *CreateReadMostlyPreAndPostAssigns( node *array_avis,
 *                                                     info *arg_info)
 *
 * @param array_avis The avis node of the array we are creating ReadMostly prfs for
 * @param arg_info
 *
 * @brief TODO
 *
 *****************************************************************************/
static void
CreateReadMostlyPreAndPostAssigns (node *array_avis, info *arg_info)
{  
    node *preassign, *postassign;

    DBUG_ENTER ();

    preassign = TBmakeAssign(
                    TBmakeLet (
                        NULL,
                        TBmakePrf(
                            F_set_read_mostly,
                            TBmakeExprs(
                                TBmakeId(array_avis), 
                                NULL))),
                    NULL
                );
    INFO_PREASSIGNS(arg_info) = TCappendAssign(INFO_PREASSIGNS(arg_info), preassign);
    
    postassign = TBmakeAssign(
                    TBmakeLet (
                        NULL,
                        TBmakePrf(
                            F_unset_read_mostly,
                            TBmakeExprs(
                                TBmakeId(array_avis), 
                                NULL))),
                    NULL
                );
    INFO_POSTASSIGNS(arg_info) = TCappendAssign(INFO_POSTASSIGNS(arg_info), postassign);

    INFO_LUT(arg_info) = LUTinsertIntoLutP(INFO_LUT(arg_info), array_avis, array_avis);
    DBUG_PRINT ( "  >>> New read_mostly primitives for buffer: %s\n", AVIS_NAME(array_avis));
    
    DBUG_RETURN ();
}

/** <!--********************************************************************-->
 * @}  <!-- Static helper functions -->
 *****************************************************************************/

/** <!--********************************************************************-->
 *
 * @name Traversal functions
 * @{
 *
 *****************************************************************************/

/** <!--********************************************************************-->
 *
 * @fn node *ICMAassign (node *arg_node, info *arg_info)
 *
 * @brief This traversal function is responsible for adding the pre- and postassigns.
 *
 * Whenever an assignment node is encountered, this function will be called. We first
 * traverse the assignment. After traversing, if we are not in a cuda with-loop, we
 * check if any new pre- and post assigns have been created. If so, this means a cuda
 * with-loop was traversed, and memory advice has to be inserted. In case we are (still)
 * in a cuda with-loop, we just traverse the next assignment.
 *       
 *****************************************************************************/
node *
ICMAassign (node *arg_node, info *arg_info)
{
    node *next;

    DBUG_ENTER();

    ASSIGN_STMT(arg_node) = TRAVdo (ASSIGN_STMT(arg_node), arg_info);

    if (!INFO_INCUDAWL(arg_info)) {
        next = ASSIGN_NEXT(arg_node);
        ASSIGN_NEXT (arg_node) = NULL;

        if (INFO_POSTASSIGNS (arg_info) != NULL) {
            DBUG_PRINT("Adding new postassigns\n");
            arg_node = TCappendAssign(arg_node, INFO_POSTASSIGNS(arg_info));
            INFO_POSTASSIGNS (arg_info) = NULL;
        }

        if (INFO_PREASSIGNS (arg_info) != NULL) {
            DBUG_PRINT("Adding new preassigns\n");
            arg_node = TCappendAssign(INFO_PREASSIGNS (arg_info), arg_node);
            INFO_PREASSIGNS (arg_info) = NULL;
        }

        node *last_assign = arg_node;
        while (ASSIGN_NEXT(last_assign) != NULL) {
            last_assign = ASSIGN_NEXT (last_assign);
        }

        ASSIGN_NEXT (last_assign) = next;
        ASSIGN_NEXT (last_assign) = TRAVopt (ASSIGN_NEXT(last_assign), arg_info);
    } else {
        ASSIGN_NEXT (arg_node) = TRAVopt(ASSIGN_NEXT(arg_node), arg_info);
    }

    DBUG_RETURN(arg_node);
}


/** <!--********************************************************************-->
 *
 * @fn node *ICMAwith (node *arg_node, info *arg_info)
 *
 * @brief The main purpose is to signal that we arrived at a cuda with-loop.
 *
 * Whenever a with-loop node is encountered, this function will be called. We first
 * check if the current node is cudarizable. If so, we generate a LUT and set the 
 * INCUDAWL flag to true, after which we start the traversal of the with-loop code.
 * If we are already in a cuda with-loop, then we encountered an inner with-loop, so
 * we keep traversing. In any other case, we stop the traversal, since we are only
 * interested in cudarizable with-loops.
 *       
 *****************************************************************************/
node *
ICMAwith (node * arg_node, info *arg_info)
{
    DBUG_ENTER();

    if (WITH_CUDARIZABLE(arg_node)){
        DBUG_PRINT("At cudarizable WL\n");
        
        INFO_LUT( arg_info ) = LUTgenerateLut ();
        INFO_INCUDAWL( arg_info ) = TRUE;

        WITH_CODE( arg_node ) = TRAVdo( WITH_CODE( arg_node ), arg_info);

        INFO_INCUDAWL( arg_info ) = FALSE;
        INFO_LUT (arg_info) = LUTremoveLut (INFO_LUT (arg_info));
    } else if (INFO_INCUDAWL(arg_info)) {
        DBUG_PRINT("At inner WL\n");

        WITH_CODE ( arg_node ) = TRAVdo(WITH_CODE (arg_node), arg_info);
    }

    DBUG_RETURN(arg_node);
}

/** <!--********************************************************************-->
 *
 * @fn node *ICMAwith2 (node *arg_node, info *arg_info)
 *
 * @brief This function ensures we traverse the code of cudarizable with2 loops.
 *
 * It seems that inner with-loops can be optimised (?) to with2 variants. In case
 * the with2 loop is cudarizable, we have to traverse its code.
 *       
 *****************************************************************************/
node *
ICMAwith2 (node * arg_node, info *arg_info)
{
    DBUG_ENTER();

    if (INFO_INCUDAWL(arg_info)) {
        DBUG_PRINT("At inner WL2 \n");

        WITH2_CODE ( arg_node ) = TRAVdo(WITH2_CODE (arg_node), arg_info);
    }

    DBUG_RETURN(arg_node);
}

/** <!--********************************************************************-->
 *
 * @fn node *ICMAprf (node *arg_node, info *arg_info)
 *
 * @brief Key traversal function that detects and traverses "idx_sel" primitives.
 *
 * To detect which arrays are read from in cudarizable with-loops, we target the
 * "idx_sel" primitive. If we encounter a prf node and we are currently traversing
 * a cudarizable with-loop, then we check what kind of primitive function it is.
 * In case we have encountered an "idx_sel", and we have no constant offset,
 * we set the corresponding flag in the
 * INFO structure to true, and traverse its second argument, which contains the
 * array ID. If we encounter any other prf, we start traversing its arguments.
 * This is necessary, because at the end of the memory phase, the idx_sel prfs are
 * wrapped in _fill_ prfs.
 *       
 *****************************************************************************/
node *
ICMAprf (node * arg_node, info *arg_info)
{
    DBUG_ENTER();

    // Only traverse if in a cuda with loop
    if (INFO_INCUDAWL (arg_info))
    {
        /**
         * We first need to check whether we arrived at F_idx_sel to signal that
         * a buffer might need to be set/unset a CUDA memory advice.
         */
        switch(PRF_PRF(arg_node)) 
        {
            case F_idx_sel:
                DBUG_PRINT ("Traversing the second argument of a prf idx_sel\n");
                INFO_INPRFSEL( arg_info ) = !PRF_CONSTANTOFFSETDIM1 (arg_node);

                PRF_ARG2( arg_node ) = TRAVdo(PRF_ARG2 (arg_node), arg_info);

                INFO_INPRFSEL( arg_info ) = FALSE;
                break;
            default:
                PRF_ARGS (arg_node) = TRAVdo (PRF_ARGS (arg_node), arg_info);
                break;
        }
    }

    DBUG_RETURN(arg_node);
}

/** <!--********************************************************************-->
 *
 * @fn node *ICMAid (node *arg_node, info *arg_info)
 *
 * @brief Creates the memory-advice pre- and postassigns when necessary.
 *
 * If we are currently in a cuda with-loop AND in a "idx_sel" primitive,
 * we need to first check whether we have already encountered this array-id
 * node. If not, we insert its avis node into the LUT and create the memory
 * advice primitives (set_read_mostly and unset_read_mostly) for this array id.
 * 
 * note: The read-mostly naming convention is used, because this is the exact
 * terminology that is used by the cuda API (cudaMemAdviseSetReadMostly). In sac,
 * these arrays are definitely read-only.
 *       
 *****************************************************************************/
node *
ICMAid (node * arg_node, info *arg_info)
{
    void **avis_lut;
    node *avis_id;

    DBUG_ENTER();

    if (INFO_INCUDAWL (arg_info) && INFO_INPRFSEL (arg_info))
    {
        avis_id  = ID_AVIS(arg_node);
        avis_lut = LUTsearchInLutP(INFO_LUT(arg_info), avis_id);

        /**
         * We check whether we encountered the buffer before by inserting
         * its avis pointer in the LUT. If not, we call a helper function,
         * which inserts the avis pointer into the LUT and creates assigns for 
         * (un)set_read_mostly primitives. These assigns are then added to the
         * INFO structure.
         */
        if (avis_lut == NULL) {
            CreateReadMostlyPreAndPostAssigns(avis_id, arg_info);
        }
    }

    DBUG_RETURN(arg_node);
}

/** <!--********************************************************************-->
 * @}  <!-- Traversal functions -->
 *****************************************************************************/

/** <!--********************************************************************-->
 *
 * @name Entry functions
 * @{
 *
 *****************************************************************************/
/** <!--********************************************************************-->
 *
 * @fn node *ICMAdoInsertMemoryAdvice (node *syntax_tree)
 *
 * @brief Entry function of the ICMA compiler sub-phase.
 * 
 *****************************************************************************/
node *
ICMAdoInsertMemoryAdvice (node *syntax_tree)
{
    info *arg_info;

    DBUG_ENTER ();

    arg_info = MakeInfo ();

    TRAVpush (TR_icma);
    syntax_tree = TRAVdo (syntax_tree, arg_info);
    TRAVpop ();

    arg_info = FreeInfo (arg_info);

    DBUG_RETURN (syntax_tree);
}

/** <!--********************************************************************-->
 * @}  <!-- Entry functions -->
 *****************************************************************************/

/** @} */
#undef DBUG_PREFIX
