/*****************************************************************************
 *
 * file:    identify_side_effect_functions.c
 *
 * prefix: DMISEFE
 *
 * description:
 *
 * We use the term side effect for functions that affect the observable world,
 * and must therefore only be executed once.
 *
 *  We mark external functions that take hidden objects as arguments as
 *  having side-effects because these are the only functions that can give 
 *  observable effects.
 *
 *  Transforms
 *    int f(int x);
 *  to
 *    // side-effects
 *    int f(int x);
 *
 * implementation:
 *
 *  We expect the traversal to start N_module. We only traverse the external
 *  function declarations. Upon entering N_fundef, we set HAS_SIDE_EFFECTS to 
 *  FALSE and traverse the arguments.
 *  If we encounter a hidden argument we set HAS_SIDE_EFFECTS and stop
 *  traversing.
 *
 *****************************************************************************/

#define DBUG_PREFIX "DMISEF"
#include "debug.h"

#include "identify_side_effect_functions.h"

#include "memory.h" /* MEM- */
#include "traverse.h" /* TRAV- */
#include "tree_basic.h" /* Son and attribute access macro's */
#include "type_utils.h" /* TUisHidden */

struct INFO {
    bool has_side_effects;
};

#define INFO_HAS_SIDE_EFFECTS(n) ((n)->has_side_effects)

static info *
MakeInfo (void)
{
    info *result;

    DBUG_ENTER ();

    result = (info *)MEMmalloc (sizeof(info));

    INFO_HAS_SIDE_EFFECTS (result) = FALSE;

    DBUG_RETURN (result);
}

static info *
FreeInfo (info *info)
{
    DBUG_ENTER ();

    info = MEMfree (info);

    DBUG_RETURN (info);
}

node *
DMISEFfundef (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    DBUG_PRINT ("Function %s is external, traversing arguments.",
            FUNDEF_NAME(arg_node));
    INFO_HAS_SIDE_EFFECTS (arg_info) = FALSE;

    /* If an argument is unique, sets INFO_HAS_SIDE_EFFECTS to TRUE. */
    FUNDEF_ARGS (arg_node) = TRAVopt (FUNDEF_ARGS(arg_node), arg_info);

    FUNDEF_DISTMEMHASSIDEEFFECTS (arg_node) = INFO_HAS_SIDE_EFFECTS (arg_info);
    DBUG_PRINT ("Function %s %s side effects.", FUNDEF_NAME (arg_node),
               (FUNDEF_DISTMEMHASSIDEEFFECTS(arg_node)) ? "has" : "has no");

    /* Continue to the next function if there is any. */
    FUNDEF_NEXT (arg_node) = TRAVopt (FUNDEF_NEXT(arg_node), arg_info);

    DBUG_RETURN (arg_node);
}

node *
DMISEFarg (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    if (TUisHidden (AVIS_TYPE (ARG_AVIS (arg_node)))) {
        INFO_HAS_SIDE_EFFECTS (arg_info) = TRUE;
        DBUG_PRINT ("Function argument is hidden");
        /* No need to traverse the rest of the arguments. */
    } else {
        DBUG_PRINT ("Function argument is not hidden");
        ARG_NEXT(arg_node) = TRAVopt (ARG_NEXT (arg_node), arg_info);
    }

    DBUG_RETURN (arg_node);
}

node *
DMISEFmodule (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    MODULE_FUNDECS (arg_node) = TRAVopt (MODULE_FUNDECS (arg_node), arg_info);

    DBUG_RETURN (arg_node);
}

node *
DMISEFdoIdentifySideEffectFunctions (node *syntax_tree)
{
    info *info;

    DBUG_ENTER ();

    DBUG_ASSERT (NODE_TYPE (syntax_tree) == N_module,
                "called with non-module node");

    TRAVpush (TR_dmisef);

    info = MakeInfo ();
    syntax_tree = TRAVdo (syntax_tree, info);
    info = FreeInfo (info);

    TRAVpop ();

    DBUG_RETURN (syntax_tree);
}

#undef DBUG_PREFIX
