/*****************************************************************************
 *
 * file:    identify_suballocated_arrays.c
 *
 * prefix: DMISAA
 *
 * description:
 *
 *  Identifies sub-allocated arrays for the distributed memory backend and marks
 *  them as not distributed (NDI).
 *
 *  Transforms
 *      int[m, n] x; // PDI
 *      int[n]    y; // PDI
 *      y = suballoc(1, x);
 * to
 *      int[m, n] x; // PDI
 *      int[n]    y; // NDI
 *      y = suballoc(1, x);
 *
 * implementation:
 *
 * In an assign, we traverse the rhs first, and set a bool in the info struct
 * if it is a suballoc. Then we traverse the lhs, and check the info struct.
 * If the rhs is a suballoc, we mark the avis as NDI.
 *
 ****************************************************************************/

#include "identify_suballocated_arrays.h"

#include "traverse.h"
#include "tree_basic.h"
#include "tree_compound.h"
#include "memory.h"
#include "new_types.h"

#define DBUG_PREFIX "DMISAA"
#include "debug.h"

/*
 * INFO structure
 */
struct INFO {
    bool isSuballoc;
};

/*
 * INFO macros
 */
#define INFO_IS_SUBALLOC(n) (n->isSuballoc)

/*
 * INFO functions
 */
static info *
MakeInfo (void)
{
    info *result;

    DBUG_ENTER ();

    result = (info *)MEMmalloc (sizeof (info));

    INFO_IS_SUBALLOC (result) = FALSE;

    DBUG_RETURN (result);
}

static info *
FreeInfo (info *info)
{
    DBUG_ENTER ();

    info = MEMfree (info);

    DBUG_RETURN (info);
}

node *
DMISAAlet (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    INFO_IS_SUBALLOC (arg_info) = (NODE_TYPE (LET_EXPR (arg_node)) == N_prf) &&
                                  (PRF_PRF (LET_EXPR (arg_node)) == F_suballoc);

    LET_IDS (arg_node) = TRAVopt (LET_IDS (arg_node), arg_info);

    INFO_IS_SUBALLOC (arg_info) = FALSE;

    LET_EXPR (arg_node) = TRAVopt (LET_EXPR (arg_node), arg_info);

    DBUG_RETURN (arg_node);
}

node *
DMISAAids (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    if (INFO_IS_SUBALLOC (arg_info)) {
        /* Mark the variable as not-distributed */
        AVIS_TYPE (IDS_AVIS (arg_node))
          = TYsetDistributed (AVIS_TYPE (IDS_AVIS (arg_node)), FALSE);
        DBUG_PRINT ("Annotated suballocated '%s' as non-distributed\n",
                            IDS_NAME (arg_node));
    }

    IDS_NEXT (arg_node) = TRAVopt (IDS_NEXT (arg_node), arg_info);

    DBUG_RETURN (arg_node);
}

node *
DMISAAdoIdentifySubAllocatedArrays (node *syntax_tree)
{
    info *info;

    DBUG_ENTER ();

    DBUG_ASSERT (NODE_TYPE (syntax_tree) == N_module,
                "called with non-module node");

    TRAVpush (TR_dmisaa);

    info = MakeInfo ();
    syntax_tree = TRAVdo (syntax_tree, info);
    info = FreeInfo (info);

    TRAVpop ();

    DBUG_RETURN (syntax_tree);
}

#undef DBUG_PREFIX
