/*****************************************************************************
 *
 * file:    mark_distmem_local_variables.c
 *
 * prefix: DMMLV
 *
 * description:
 *
 * Any variable that is local to a with-loop, can never be distributed. We
 * mark them as such.
 *
 *  Transforms
 *    int[*] x; // PDI
 *    int[*] y; // PDI
 *    x = with {
 *          ...
 *          y = with {
 *              ...
 *          }
 *        }
 *  to
 *    int[*] x; // PDI
 *    int[*] y; // NDI
 *    x = with {
 *          ...
 *          y = with {
 *              ...
 *          }
 *        }
 *
 * implementation:
 *
 *  We keep track of whether we are in a with-loop. Whenever we encounter a
 *  N_ids inside such a with-loop, we mark it as non-distributed.
 *
 ****************************************************************************/

#define DBUG_PREFIX "DMMLV"
#include "debug.h"

#include "traverse.h" /* TRAV- */
#include "tree_basic.h" /* Son and attribute access macro's */
#include "tree_compound.h" /* IDS_NAME */
#include "memory.h" /* MEM- */
#include "new_types.h" /* TY- */

struct INFO {
    bool inWith;
};

#define INFO_IN_WITH(n) (n->inWith)

static info *
MakeInfo (void)
{
    info *result;

    DBUG_ENTER();

    result = (info *)MEMmalloc (sizeof(info));

    INFO_IN_WITH (result) = FALSE;

    DBUG_RETURN (result);
}

static info *
FreeInfo (info *info)
{
    DBUG_ENTER();

    info = MEMfree (info);

    DBUG_RETURN (info);
}

node *
DMMLVids (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    if (INFO_IN_WITH (arg_info)) {
        AVIS_TYPE (IDS_AVIS (arg_node))
          = TYsetDistributed (AVIS_TYPE (IDS_AVIS (arg_node)), FALSE);
        DBUG_PRINT("Annotated '%s' as non-distributable\n",
                        IDS_NAME (arg_node));
    }

    /* Continue the assignment chain */
    IDS_NEXT (arg_node) = TRAVopt (IDS_NEXT (arg_node), arg_info);

    DBUG_RETURN (arg_node);
}

node *
DMMLVwith2 (node *arg_node, info *arg_info)
{
    DBUG_ENTER();

    bool already_in_with = INFO_IN_WITH (arg_info);

    INFO_IN_WITH (arg_info) = TRUE;

    arg_node = TRAVcont (arg_node, arg_info);

    INFO_IN_WITH (arg_info) = already_in_with;

    DBUG_RETURN (arg_node);
}

node *
DMMLVdoMarkLocalVariables (node *syntax_tree)
{
    info *info;

    DBUG_ENTER ();

    DBUG_ASSERT (NODE_TYPE (syntax_tree) == N_module, 
                "called with non-module node");

    TRAVpush (TR_dmmlv);

    info = MakeInfo ();
    syntax_tree = TRAVdo (syntax_tree, info);
    info = FreeInfo (info);

    TRAVpop ();

    DBUG_RETURN (syntax_tree);
}

#undef DBUG_PREFIX
