/*****************************************************************************
 *
 * file: mark_distributed_fold.c
 *
 * prefix: DMMDF
 *
 * description:
 *
 *  We mark all (potentially) top-level fold-loops as potentially parallel, 
 *  as long as the foldfun is not external. This is to simplify the code 
 *  generation, as foldfun with non-trivial linksign are hard to deal with.
 *
 *  TODO (Optimisation): don't mark small fold-loops. This is non-trivial
 *  as the size of the loop is hidden behind multiple N_wlseg, N_grid,
 *  N_wlstride nodes.
 *
 *  Transforms
 *     y = with {
 *              ...
 *         }: fold(+, 0
 *  to
 *     y = with { // Distmem: parallel fold
 *              ...
 *         }: fold(+, 0);
 *
 *  but leaves
 *     x = {iv -> with {...}: fold(+, 0)};
 *  alone.
 *
 * implementation:
 *  Mark fold-loops (in the flag WITH2_DISTFOLD) 
 *  and don't traverse the body of with-loops.
 *
 ****************************************************************************/

#include "mark_distributed_fold.h"

#define DBUG_PREFIX "DMMDF"
#include "debug.h"

#include "traverse.h" /* TRAV- */
#include "tree_basic.h" /* Attribute accessors */

node *
DMMDFwith2 (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    WITH2_DISTFOLD (arg_node) = (NODE_TYPE (WITH2_WITHOP (arg_node)) == N_fold)
                   && !FUNDEF_ISEXTERN (FOLD_FUNDEF (WITH2_WITHOP (arg_node)));

    DBUG_PRINT ("Marked fold-loop %s as distributed.",
            (WITH2_MEMID (arg_node) == NULL) ?
              "<name unknown>" : AVIS_NAME (ID_AVIS (WITH2_MEMID (arg_node))));

    /* Stop traversing because we don't want to mark nested with-loops. */

    DBUG_RETURN (arg_node);
}

node *
DMMDFdoMarkDistributedFold (node *syntax_tree)
{
    DBUG_ENTER ();

    DBUG_ASSERT (NODE_TYPE (syntax_tree) == N_module,
                "called with non-module node");

    TRAVpush (TR_dmmdf);

    syntax_tree = TRAVdo (syntax_tree, NULL);

    TRAVpop ();

    DBUG_RETURN (syntax_tree);
}

#undef DBUG_PREFIX
