/*****************************************************************************
 *
 * file:    mark_fold_result.c
 *
 * prefix: DMMFR
 *
 * description:
 *
 *  Marks the result of a fold with-loop as non-distributed. 
 *
 *  The neutral element is potentially distributed. We do not have to deal
 *  with this, because the implementation of the backend makes sure that
 *  assigning a PDI to NDI makes a deep copy.
 *
 *  Transforms
 *    int[10] y; // PDI
 *    y = with {...} fold ...
 *  to
 *    int[10] y; // NDI
 *    y = with {...} fold ...
 *
 * implementation:
 *
 *  We check at the let node whether the right-hand side is a fold, and mark
 *  this decision in our info struct. When then traversing the lhs, the ids are
 *  marked accordingly.
 *
 ****************************************************************************/

#define DBUG_PREFIX "DMMFR"
#include "debug.h"

#include "traverse.h" /* TRAV- */
#include "tree_basic.h" /* Son and attribute access macro's */
#include "tree_compound.h" /* IDS_NAME */
#include "memory.h" /* MEM- */
#include "new_types.h" /* TY- */

struct INFO {
    bool isFold;
};

#define INFO_ISFOLD(n) (n->isFold)

static info *
MakeInfo (void)
{
    info *result;

    DBUG_ENTER();

    result = (info *)MEMmalloc (sizeof(info));

    INFO_ISFOLD (result) = FALSE;

    DBUG_RETURN (result);
}

static info *
FreeInfo (info *info)
{
    DBUG_ENTER();

    info = MEMfree (info);

    DBUG_RETURN (info);
}

node *
DMMFRlet (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    node *rhs = LET_EXPR (arg_node);
    INFO_ISFOLD (arg_info) = (NODE_TYPE (rhs) == N_with2) &&
            (NODE_TYPE (WITH2_WITHOP (rhs)) == N_fold);

    LET_IDS (arg_node) = TRAVopt (LET_IDS (arg_node), arg_info);

    /* Everything insde a with-loops is marked as non-distributed by DMMNW,
     * so with-loops on the rhs do not have to be traversed. There are no other
     * node types on the rhs that can contain other let nodes. So the rhs
     * does not have to be traversed. */

    DBUG_RETURN (arg_node);
}

node *
DMMFRids (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    if (INFO_ISFOLD (arg_info)) {
        AVIS_TYPE (IDS_AVIS (arg_node))
          = TYsetDistributed (AVIS_TYPE (IDS_AVIS (arg_node)), FALSE);
        DBUG_PRINT("Annotated '%s' as nondistributable\n",
                IDS_NAME (arg_node));
    }


    IDS_NEXT (arg_node) = TRAVopt (IDS_NEXT (arg_node), arg_info);

    DBUG_RETURN (arg_node);
}

/*
 * Traversal start function
 */
node *
DMMFRdoMarkFoldResult (node *syntax_tree)
{
    info *info;

    DBUG_ENTER ();

    DBUG_ASSERT (NODE_TYPE (syntax_tree) == N_module,
                "called with non-module node");

    TRAVpush (TR_dmmfr);

    info = MakeInfo ();
    syntax_tree = TRAVdo (syntax_tree, info);
    info = FreeInfo (info);

    TRAVpop ();

    DBUG_RETURN (syntax_tree);
}

#undef DBUG_PREFIX
