/*****************************************************************************
 *
 * file:    mark_potentially_distributed.c
 *
 * prefix: DMMPD
 *
 * description:
 *
 *  By default, all variables have been marked as non-distributable (NDI).
 *  This pass marks everything that may potentially be distributed as PDI
 *  (aka everything that is not a scalar, and not hidden).
 *  This is necessary for correctness. Later phases () will change some
 *  back to NDI. This is only for performance, it saves a runtime check.
 *
 *  Transforms
 *      int[*]     x; // NDI (default)
 *      int        y; // NDI (default)
 *      string[10] z; // NDI (default)
 *  to
 *      int[*]     x; // PDI
 *      int        y; // NDI
 *      string[10] z; // NDI
 *
 * implementation:
 *
 *   We traverse the avis node and query the type information.
 *
 ****************************************************************************/

#include "mark_non_scalars.h"

#define DBUG_PREFIX "DMMNS"
#include "debug.h"

#include "traverse.h"   /* TRAV- */
#include "tree_basic.h" /* AVIS_TYPE */
#include "new_types.h"  /* TY- */
#include "type_utils.h" /* TU- */

node *
DMMNSavis (node *arg_node, info *arg_info)
{
    DBUG_ENTER();

    bool is_scalar = TUisScalar (AVIS_TYPE (arg_node));
    bool is_hidden = TUisHidden (AVIS_TYPE (arg_node));
    AVIS_TYPE (arg_node) = TYsetDistributed (AVIS_TYPE (arg_node), 
                                             !is_scalar && !is_hidden);
    DBUG_PRINT ("Marked array %s as %sdistributable.",
            AVIS_NAME (arg_node), (is_scalar || is_hidden) ? "non-" : "");

    DBUG_RETURN (arg_node);
}

node *
DMMNSdoMarkNonScalars (node *syntax_tree)
{
    DBUG_ENTER ();

    DBUG_ASSERT (NODE_TYPE (syntax_tree) == N_module,
                "called with non-module node");

    TRAVpush (TR_dmmns);

    syntax_tree = TRAVdo (syntax_tree, NULL);

    TRAVpop ();

    DBUG_RETURN (syntax_tree);
}

#undef DBUG_PREFIX
