/*****************************************************************************
 *
 * file:    mark_prf_result.c
 *
 * prefix: DMMPR
 *
 * description:
 *
 *  Marks the lhs of primitive functions that compute a new array as NDI.
 *  The reason for this is that these primitive functions are implemented
 *  directly and we do not generate distributed code for them. The exception
 *  to this is reshape and variations.
 *
 *  Transforms
 *      int[*] x; // PDI
 *      int[*] y; // PDI
 *      x = _prf...
 *      y = _reshape(...)
 *  to
 *      int[*] x; // NDI
 *      int[*] y; // PDI
 *      x = _prf...
 *      y = _reshape(...)
 *
 ****************************************************************************/

#define DBUG_PREFIX "DMMPR"
#include "debug.h"

#include "traverse.h"   /* TRAV- */
#include "tree_basic.h" /* Son and attribute access macro's */
#include "tree_compound.h" /* IDS_NAME */
#include "memory.h"    /* MEM- */
#include "new_types.h" /* TY- */

struct INFO {
    bool isPrf;
};

#define INFO_ISPRF(n) ((n)->isPrf)

static info *
MakeInfo (void)
{
    info *result;

    DBUG_ENTER();

    result = (info *)MEMmalloc (sizeof(info));

    INFO_ISPRF (result) = FALSE;

    DBUG_RETURN (result);
}

static info *
FreeInfo (info *info)
{
    DBUG_ENTER();

    info = MEMfree (info);

    DBUG_RETURN (info);
}

node *
DMMPRlet (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    node *rhs = LET_EXPR (arg_node);
    if (NODE_TYPE (rhs) == N_prf) {
        int op = PRF_PRF (rhs);
        INFO_ISPRF (arg_info) = // TODO: !(op == F_reshape)
              (op == F_add_VxV) || (op == F_add_VxS)  || (op == F_add_SxV) ||
              (op == F_sub_VxV) || (op == F_sub_VxS)  || (op == F_sub_SxV) ||
              (op == F_mul_VxV) || (op == F_mul_VxS)  || (op == F_mul_SxV) ||
              (op == F_div_VxV) || (op == F_div_VxS)  || (op == F_div_SxV) ||
              (op == F_mod_VxV) || (op == F_mod_VxS)  || (op == F_mod_SxV) ||
              (op == F_aplmod_VxV) || (op == F_aplmod_VxS)
                     || (op == F_aplmod_SxV) ||
              (op == F_aplmod_VxV) || (op == F_aplmod_VxS)
                     || (op == F_aplmod_SxV) ||
              (op == F_neg_V)   || (op == F_abs_V)    ||
              (op == F_min_VxV) || (op == F_min_VxS)  || (op == F_min_SxV) ||
              (op == F_max_VxV) || (op == F_max_VxS)  || (op == F_max_SxV) ||
              (op == F_eq_VxV)  || (op == F_eq_VxS)   || (op == F_eq_SxV) ||
              (op == F_neq_VxV) || (op == F_neq_VxS)  || (op == F_neq_SxV) ||
              (op == F_le_VxV)  || (op == F_le_VxS)   || (op == F_le_SxV) ||
              (op == F_lt_VxV)  || (op == F_lt_VxS)   || (op == F_lt_SxV) ||
              (op == F_ge_VxV)  || (op == F_ge_VxS)   || (op == F_ge_SxV) ||
              (op == F_gt_VxV)  || (op == F_gt_VxS)   || (op == F_gt_SxV) ||
              (op == F_cat_VxV) || (op == F_take_SxV) || (op == F_drop_SxV);
    }

    LET_IDS (arg_node) = TRAVopt (LET_IDS (arg_node), arg_info);

    INFO_ISPRF (arg_info) = FALSE;

    DBUG_RETURN (arg_node);
}

node *
DMMPRids (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    if (INFO_ISPRF (arg_info)) {
        AVIS_TYPE (IDS_AVIS (arg_node))
          = TYsetDistributed (AVIS_TYPE (IDS_AVIS (arg_node)), FALSE);
        DBUG_PRINT("Annotated '%s' as nondistributable\n",
                IDS_NAME (arg_node));
    }

    IDS_NEXT (arg_node) = TRAVopt (IDS_NEXT (arg_node), arg_info);

    DBUG_RETURN (arg_node);
}

node *
DMMPRdoMarkPrfResult (node *syntax_tree)
{
    info *info;

    DBUG_ENTER ();

    DBUG_ASSERT (NODE_TYPE (syntax_tree) == N_module,
                "called with non-module node");

    TRAVpush (TR_dmmpr);

    info = MakeInfo ();
    syntax_tree = TRAVdo (syntax_tree, info);
    info = FreeInfo (info);

    TRAVpop ();

    DBUG_RETURN (syntax_tree);
}

#undef DBUG_PREFIX
