/*****************************************************************************
 *
 * file:    mark_small_arrays.c
 *
 * prefix: DMMSA
 *
 * IMPORTANT:
 *
 *  This must be run after DMMNS.
 *
 * description:
 *
 *  This is an optimisation for the distributed memory backend. We mark arrays
 *  that are compile-time known to be smaller than distmem_min_elems_per_node 
 *  as non-distributable (NDI). This saves checks at runtime. 
 *
 *  Tranforms
 *      int[3] x; // PDI
 *  to
 *      int[3] x; // NDI
 *
 * implementation:
 *
 *  We traverse all avis nodes, and set the distributed type information to ndi
 *  in case the array is too small.
 *
 ****************************************************************************/

#include "mark_small_arrays.h"

#define DBUG_PREFIX "DMMSA"
#include "debug.h"

#include "globals.h" /* global.distmem_min_elems_per_node */
#include "traverse.h" /* TRAV- */
#include "tree_basic.h" /* AVIS_TYPE, global */
#include "new_types.h" /* TY- */
#include "type_utils.h" /* TUshapeKnown */
#include "shape.h" /* SHgetUnrLen */

node *
DMMSAavis (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    if (TUshapeKnown (AVIS_TYPE (arg_node))) {
        long long size = SHgetUnrLen (TYgetShape (AVIS_TYPE (arg_node)));
        /* We always use at least two nodes, so in this case the array can
         * never be distributed. */
        if (size < 2 * global.distmem_min_elems_per_node) {
            AVIS_TYPE (arg_node) = TYsetDistributed (AVIS_TYPE(arg_node),
                                                     FALSE);
            DBUG_PRINT ("Marked %s of %lld elements as non-distributable "
                        "because it is too small.",
                        AVIS_NAME (arg_node), size);
        }
    }

    DBUG_RETURN (arg_node);
}

node *
DMMSAdoMarkSmallArrays (node *syntax_tree)
{
    DBUG_ENTER ();

    DBUG_ASSERT (NODE_TYPE (syntax_tree) == N_module,
                "called with non-module node");

    TRAVpush (TR_dmmsa);

    syntax_tree = TRAVdo (syntax_tree, NULL);

    TRAVpop();

    DBUG_RETURN (syntax_tree);
}

#undef DBUG_PREFIX
