#include "config.h"
#include "fun-attrs.h"

#ifdef HAVE_GETTIME

#include <string.h>
#include <time.h>

#include "ctinfo.h"
#include "globals.h"
#include "LookUpTable.h"
#include "memory.h"
#include "phase_info.h"
#include "tree_basic.h"
#include "types.h"

#define DBUG_PREFIX "TIME"
#include "debug.h"

#include "profiler.h"

static lut_t *timetable = 0;
static FILE *reportfile = 0;

static const char *phasename[] = {
    "",
    "Phase",
    "Subphase",
    "Cycle",
    "Cycle (fun)",
    "Cyclephase",
    "Cyclephase (fun)"
};

static bool
ShouldTime (void)
{
    phase_type_t phasetype = PHIphaseType (global.compiler_anyphase);
    return (int)phasetype > 0 && (int)phasetype <= global.timefreq;
}

/******************************************************************************
 *
 * @fn void CreateReport (timeinfo_t *phasetime)
 *
 * @brief Generate an entry for this phasetime. The format is described at:
 * https://docs.google.com/document/d/1CvAClvFfyA5R-PhYUmn5OOQtYMH4h6I0nSsKchNAySU/edit?tab=t.0
 *
 ******************************************************************************/
static void
CreateReport (timeinfo_t *phasetime)
{
    phase_type_t phase = PHIphaseType (phasetime->phase);
    const char *phase_str = PHIphaseText (phasetime->phase);
    ssize_t start = ((ssize_t)phasetime->start.tv_sec * (ssize_t)1e9
                     + phasetime->start.tv_nsec) / 1000;
    ssize_t end = ((ssize_t)phasetime->end.tv_sec * (ssize_t)1e9
                   + phasetime->end.tv_nsec) / 1000;

    if (reportfile == NULL) {
        char *filename;
        STRasprintf (&filename, "%s.treport", global.outfilename);
        reportfile = fopen (filename, "w");
    }

    fprintf (reportfile,
             "{\"cat\":\"%s\",\"name\":\"%s\",\"ts\":%ld,\"ph\":\"B\"},\n",
             phasename[phase], phase_str, start);

    fprintf (reportfile,
             "{\"cat\":\"%s\",\"name\":\"%s\",\"ts\":%ld,\"ph\":\"E\"},\n",
             phasename[phase], phase_str, end);
}

static void *
WrapCreateReport (void *head, void *node, void *Key)
{
    while (node != NULL) {
        CreateReport ((timeinfo_t *)node);
        node = ((timeinfo_t *)node)->next;
    }

    return node;
}

void
TIMEbegin (compiler_phase_t phase)
{
    if (!ShouldTime ()) {
        return;
    }

    if (timetable == NULL) {
        timetable = LUTgenerateLut ();
    }

    timeinfo_t *prev = NULL;
    if ((prev = (void *)LUTsearchInLutP (timetable, (void *)phase))) {
        prev = *((void **)prev);
    }

    timeinfo_t *phasetime = MEMmalloc (sizeof (timeinfo_t));
    memset (phasetime, 0, sizeof (timeinfo_t));
    phasetime->phase = phase;
    clock_gettime (CLOCK_MONOTONIC, &phasetime->start);
    phasetime->next = prev;

    void *founditem = NULL;
    timetable = LUTupdateLutP (timetable, (void *)phasetime->phase,
                               phasetime, &founditem);
}

void
TIMEend (compiler_phase_t phase)
{
    timeinfo_t *phasetime;

    if (!(phasetime = (void *)LUTsearchInLutP (timetable, (void *)phase))) {
        return;
    } else {
        phasetime = *((void **)phasetime);
    }

    clock_gettime (CLOCK_MONOTONIC, &phasetime->end);
}

node *
TIMEtimeReport (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    char *filename = MEMmalloc ((strlen (global.outfilename) + 8)
                                * sizeof (char));
    sprintf (filename, "%s.treport", global.outfilename);
    reportfile = fopen (filename, "w");

    fprintf (reportfile, "{\"traceEvents\":[\n");

    LUTfoldLutP (timetable, NULL, WrapCreateReport);

    fprintf (reportfile, "{\"name\":\"dummy\",\"ph\":\"M\"}]}\n");

    DBUG_RETURN (arg_node);
}

#else

static UNUSED int dummy_variable; /* Silences empty file warning */

#endif /* HAVE_GETTIME */
