/**<!--******************************************************************-->
 *
 * @file infer_constant_access_offset.c
 *
 * In the multigpu backend, we want to give the unified memory
 * a hint cudaMemAdviseSetMostly whenever an array within a with-loop
 * is accessed by all nodes. If it is (mainly) accessed by one GPU, this gives
 * unnecessary overhead. This phase marks selections that are almost always
 * local. This information is then used in ICMA.
 *
 * Arrays are distributed over the first axis, so if the access is
 *  A[[jv[0] + c, ...]] for jv the index variable of the outer-most with-loop,
 * and c small, the accesses are mostly local.
 *
 * We set OffsetConstantDim1 in the _sel_ node. If this is changed into
 * an idx_sel node by ive_split_selections.c, we copy over the flag.
 *
 * Implementation:
 *  Whenever we traverse an _sel_VxA_(iv, x) node, we pattern match on
 *  iv. To do this, we need to have the index vector jv=[j1, ...] of the
 *  top-level with-loop in our info struct. We set it to NULL whenever
 *  we are not in a with-loop.
 *
 *  We keep track of whether we are in a with-loop or not. If we encounter
 *  a with-loop while we are already in one, we simply traverse the code.
 *
 *  If we encounter a with-loop for the first time we have to be a bit careful:
 *  The standard traversal order is
 *     "partition1, ..., partitionn, code1, ..., coden"
 *  We want
 *     "partition1, code1, ..., partitionn, coden"
 *
 *  So we do NOT traverse the code son of N_with, but instead traverse this
 *  through the code attribute of N_part.
 *
 *  If codes are shared between multiple partitions we traverse it multiple
 *  times, but the offset annotation is the same, so this is not a problem.
 *
 **************************************************************************/

#define DBUG_PREFIX "ICAO"

#include "infer_constant_access_offset.h"

#include "constants.h"
#include "debug.h"
#include "memory.h"
#include "pattern_match.h"
#include "traverse.h"
#include "tree_basic.h"
#include "types.h"

struct INFO {
    bool topWith;
    node *j1;
    node *jv;
};

#define INFO_INWITH(n) ((n)->topWith)
#define INFO_J1(n) ((n)->j1)
#define INFO_JV(n) ((n)->jv)

static info *
MakeInfo (void)
{
    info *result;

    DBUG_ENTER ();

    result = (info *)MEMmalloc (sizeof (info));

    INFO_INWITH (result) = FALSE;
    INFO_J1 (result) = NULL;
    INFO_JV (result) = NULL;

    DBUG_RETURN (result);
}

static info *
FreeInfo (info *info)
{
    DBUG_ENTER ();

    info = MEMfree (info);

    DBUG_RETURN (info);
}

node *ICAOwith (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    bool already_in_with = INFO_INWITH (arg_info);

    if (already_in_with) {
        WITH_CODE (arg_node) = TRAVdo (WITH_CODE (arg_node), arg_info);
    } else {
        INFO_INWITH (arg_info) = TRUE;
        WITH_PART (arg_node) = TRAVdo (WITH_PART (arg_node), arg_info);
        INFO_INWITH (arg_info) = FALSE;
        INFO_JV (arg_info) = NULL;
        INFO_J1 (arg_info) = NULL;
    }

    DBUG_RETURN (arg_node);
}

node *ICAOpart (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    PART_WITHID (arg_node) = TRAVdo (PART_WITHID (arg_node), arg_info);
    /* We do not (de)allocate memory, so this will modify the same
     * memory as the code son of the parent N_with node. */
    PART_CODE (arg_node) = TRAVdo (PART_CODE (arg_node), arg_info);

    DBUG_RETURN (arg_node);
}

node *ICAOwithid (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    /* WITHID_IDS may not exist, in case of AUD arrays */
    node *ids = WITHID_IDS (arg_node);
    DBUG_ASSERT ((ids == NULL) || (NODE_TYPE (ids) == N_ids),
                 "In ptc_ivd - mem_alloc, withid_ids should be N_ids");
    INFO_J1 (arg_info) = (ids == NULL) ? NULL : IDS_AVIS (ids);

    node *vec = WITHID_VEC (arg_node);
    DBUG_ASSERT (vec != NULL,
                 "In ptc_ivd - mem_alloc, withid_vec should exist");
    DBUG_ASSERT (NODE_TYPE (vec) == N_ids,
                 "In ptc_ivd - mem_alloc, withid_vec should be N_ids");
    INFO_JV (arg_info) = IDS_AVIS (vec);

    DBUG_RETURN (arg_node);
}

node *ICAOprf (node *arg_node, info *arg_info)
{
    /**
     * Variable naming convention: index vector of top-level with-loop is
     *  jv = [j1, ...].
     * Selection is _sel_VxA(iv, x)
     **/

    DBUG_ENTER ();

    node *jv = INFO_JV (arg_info);
    node *j1 = INFO_J1 (arg_info);
    bool in_with = INFO_INWITH (arg_info);
    DBUG_ASSERT (in_with == (jv != NULL),
            "jv must be set iff we are in a with-loop");

    if (!in_with || PRF_PRF (arg_node) != F_sel_VxA) {
        goto continue_traverse;
    }

    DBUG_PRINT("jv = %s", AVIS_NAME (jv));
    DBUG_ASSERT (NODE_TYPE (jv) == N_avis, "jv must be an avis");

    node *iv = EXPRS_EXPR (PRF_ARGS (arg_node));
    DBUG_ASSERT (NODE_TYPE (iv) == N_id,
                "First argument of sel_VxA should be N_id");

    /* iv = jv */
    pattern *pat_map = PMparam(1, PMAhasAvis (&jv));
    if (PMmatchFlat (pat_map, iv)) {
        PRF_CONSTANTOFFSETDIM1 (arg_node)    = TRUE;
        PRF_CONSTANTOFFSETNUM (arg_node) = 0;
        DBUG_PRINT ("Matched map iv = jv");
    }
    PMfree (pat_map);

    /* AUD parent with-loop, nothing more to do */
    if (j1 == NULL) {
        goto continue_traverse;
    }
    DBUG_PRINT("j1 = %s", AVIS_NAME (j1));

    node *i1;
    pattern *pat_i1 = PMarray (0, 1, PMskip (1, PMAgetNode (&i1)));
    if (!PMmatchFlat (pat_i1, iv)) {
        goto cleanup_i1;
    }
    DBUG_PRINT ("Matched i1 = %s", AVIS_NAME (ID_AVIS (EXPRS_EXPR (i1))));
    constant *con = NULL;

    /* iv = [j1 + c, ...] */
    pattern *pat_off1 = PMprf (1, PMAisPrf (F_add_SxS),
                               2, PMparam (1, PMAhasAvis (&j1)),
                                  PMconst (1, PMAgetVal (&con)));
    if (PMmatchFlat (pat_off1, i1)) {
        PRF_CONSTANTOFFSETDIM1 (arg_node)    = TRUE;
        PRF_CONSTANTOFFSETNUM (arg_node) = COconst2IntAndFree (con);
        DBUG_PRINT ("Matched iv = [j1 + %d, ...]",
                    PRF_CONSTANTOFFSETNUM (arg_node));
    }
    PMfree (pat_off1);

    /* iv = [c + j1, ...] */
    pattern *pat_off2 = PMprf (1, PMAisPrf (F_add_SxS),
                               2, PMconst (1, PMAgetVal (&con)),
                                  PMparam (1, PMAhasAvis (&j1)));
    if (PMmatchFlat (pat_off2, i1)) {
        PRF_CONSTANTOFFSETDIM1 (arg_node)    = TRUE;
        PRF_CONSTANTOFFSETNUM (arg_node) = COconst2IntAndFree (con);
        DBUG_PRINT ("Matched iv = [%d + j1, ...]",
                    PRF_CONSTANTOFFSETNUM (arg_node));
    }
    PMfree (pat_off2);

    /* iv = [j1, ...] */
    pattern *pat_off3 = PMparam (1, PMAhasAvis (&j1));
    if (PMmatchFlat (pat_off3, i1)) {
        PRF_CONSTANTOFFSETDIM1 (arg_node)    = TRUE;
        PRF_CONSTANTOFFSETNUM (arg_node) = 0;
        DBUG_PRINT ("Matched iv = [j1, ...]");
    }
    PMfree (pat_off3);

    /* iv = [j1 - c, ...] */
    pattern *pat_off4 = PMprf (1, PMAisPrf (F_sub_SxS),
                               2, PMparam (1, PMAhasAvis (&j1)),
                                  PMconst (1, PMAgetVal (&con)));
    if (PMmatchFlat (pat_off4, i1)) {
        PRF_CONSTANTOFFSETDIM1 (arg_node)    = TRUE;
        PRF_CONSTANTOFFSETNUM (arg_node) = -COconst2IntAndFree (con);
        DBUG_PRINT ("Matched iv = [j1 - %d, ...]",
                    -PRF_CONSTANTOFFSETNUM (arg_node));
    }
    PMfree (pat_off4);

    /* iv = [c - j1, ...] */
    pattern *pat_off5 = PMprf (1, PMAisPrf (F_sub_SxS),
                               2, PMconst (1, PMAgetVal (&con)),
                                  PMparam (1, PMAhasAvis (&j1)));
    if (PMmatchFlat (pat_off5, i1)) {
        PRF_CONSTANTOFFSETDIM1 (arg_node)    = TRUE;
        PRF_CONSTANTOFFSETNUM (arg_node) = -COconst2IntAndFree (con);
        DBUG_PRINT ("Matched iv = [%d - j1, ...]",
                    -PRF_CONSTANTOFFSETNUM (arg_node));
    }
    PMfree (pat_off5);

cleanup_i1:
    PMfree (pat_i1);

continue_traverse:
    PRF_ARGS (arg_node) = TRAVopt (PRF_ARGS (arg_node), arg_info);

    DBUG_RETURN (arg_node);
}

node *ICAOdoInferAccessOffset(node *syntax_tree)
{
    info *info;
    DBUG_ENTER ();

    info = MakeInfo ();

    TRAVpush (TR_icao);
    syntax_tree = TRAVdo (syntax_tree, info);
    TRAVpop ();

    info = FreeInfo (info);

    DBUG_RETURN (syntax_tree);
}
