/*****************************************************************************
 *
 * file:   distmem.h
 *
 * prefix: SAC_DISTMEM_ (no tracing) / SAC_DISTMEM_TR_ (tracing)
 *         SAC_DISTMEM_PR_ (profiling)
 *
 * description: The header file declares the interface of libsacdistmem.
 *              From outside libsacdistmem use
 *              the provided macros and do not call the functions directly.
 *              This way you do not need to check whether the distributed
 *              memory backend is actually used (for some functions) and
 *              whether the tracing/profiling version needs to be used.
 *
 *****************************************************************************/

#ifndef _SAC_DISTMEM_H_
#define _SAC_DISTMEM_H_

#ifndef SAC_C_EXTERN_VAR
#define SAC_C_EXTERN_VAR extern
#endif

/******************************************
 * Things that need to be defined / declared
 * because they are included in the profiling
 * macros of rt_profile.h
 *
 * FIXME (ugly to define it in cases where
 * we do not use the backend).
 ******************************************/
#define SAC_DISTMEM_RANK_UNDEFINED UINT_MAX
#define SAC_DISTMEM_TRACE_PROFILE_RANK_ANY UINT_MAX

/* Rank of this node */
SAC_C_EXTERN_VAR unsigned int SAC_DISTMEM_rank;

/* If not equal to SAC_DISTMEM_TRACE_PROFILE_RANK_ANY, only produce
 * trace output for this rank.
 * We always declare this variable for tracing purposes. */
SAC_C_EXTERN_VAR unsigned int SAC_DISTMEM_trace_profile_rank;

#if SAC_DO_DISTMEM

#include <stdlib.h>
#include <inttypes.h>
#include <limits.h>
#include "runtime/essentials_h/bool.h"

#if SAC_DO_TRACE_DISTMEM
    #define SHRAY_DEBUG
#elif SAC_DO_PROFILE_DISTMEM
    #define SHRAY_PROFILE
#endif
#include "shray.h"

/* Source node. Only this nodes executes functions with side-effects. */
#define SAC_DISTMEM_RANK_SOURCE 0

/* Type for the execution mode */
typedef enum SAC_DISTMEM_exec_mode_enum {
    /* All nodes are executing the same code. */
    SAC_DISTMEM_exec_mode_replicated,
    /* Each node is executing its share of work. */
    SAC_DISTMEM_exec_mode_dist,
    /* Each node is executing its share of a fold loop. */
    SAC_DISTMEM_exec_mode_fold,
} SAC_DISTMEM_exec_mode_t;

/******************************************
 * Global variables
 *******************************************/

/* Number of nodes */
SAC_C_EXTERN_VAR unsigned int SAC_DISTMEM_size;

/* Current execution mode. */
SAC_C_EXTERN_VAR SAC_DISTMEM_exec_mode_t SAC_DISTMEM_exec_mode;

/* Minimum number of array elements per node such that an array gets
 * distributed */
SAC_C_EXTERN_VAR size_t SAC_DISTMEM_min_elems_per_node;

/******************************************
 * Global variables used for
 * tracing/profiling
 *******************************************/

/* Number of segfaults = page fetches */
SAC_C_EXTERN_VAR unsigned long SAC_DISTMEM_TR_num_segfaults;

/* Number of barriers */
SAC_C_EXTERN_VAR unsigned long SAC_DISTMEM_TR_num_barriers;

/******************************************
 * No tracing declarations
 *******************************************/

/** <!--********************************************************************-->
 *
 * @fn size_t SAC_DISTMEM_DetDoDistributeArray(size_t total_elems,
 *                                             size_t dim0_size)
 *
 *   @brief   Determines whether an array of the given size should be
 *            distributed.
 *
 *   @param total_elems   total number of array elements
 *   @param dim0_size     size of the first dimension of the array
 *   @return              TRUE iff the array should be distributed,
 *
 ******************************************************************************/

bool SAC_DISTMEM_DetDoDistrArr(size_t total_elems, size_t dim0_size);

/** <!--********************************************************************-->
 *
 * @fn void SAC_COMMLIB_Broadcast( void *buffer, size_t size, int root)
 *
 *   @brief   Broadcasts buffer from node root to all other nodes.
 *
 *   @param buffer        Start of the data to be broadcast. Other nodes need to
 *                        have allocated memory here as well.
 *   @param size          Size of the buffer
 *   @param root          Node that does the broadcasting
 *
 ******************************************************************************/

void SAC_COMMLIB_Broadcast(void *buffer, size_t size, int root);

/** <!--********************************************************************-->
 *
 * @fn SAC_DISTMEM_Allgather(void *dest, void *src, size_t size)
 *
 *   @brief  Gathers src from all nodes into dest, on every node.
 *
 *
 *   @param dest     array of size 'size * SAC_DISTMEM_size'
 *          src      object to be sent
 *          size     size of object at src
 *
 *
 ******************************************************************************/

void SAC_DISTMEM_Allgather(void *dest, void *src, size_t size);

void SAC_DISTMEM_Allgatherv(size_t p, void *dest, void *src, size_t *sizes);

void SAC_DISTMEM_Barrier(void);

/******************************************
 * Tracing declarations
 *******************************************/

bool SAC_DISTMEM_TR_DetDoDistrArr(size_t total_elems, size_t dim0_size);

void SAC_COMMLIB_TR_Broadcast(void *buffer, size_t size, int root);

void SAC_DISTMEM_TR_Allgather(void *dest, void *src, size_t size);

void SAC_DISTMEM_TR_Allgatherv(size_t p, void *dest, void *src, size_t *sizes);

void SAC_DISTMEM_TR_Barrier(void);

/******************************************
 * Profiling declarations
 *******************************************/

bool SAC_DISTMEM_PR_DetDoDistrArr (size_t total_elems, size_t dim0_size);

void SAC_COMMLIB_PR_Broadcast(void *buffer, size_t size, int root);

void SAC_DISTMEM_PR_Allgather(void *dest, void *src, size_t size);

void SAC_DISTMEM_PR_Allgatherv(size_t p, void *dest, void *src, size_t *sizes);

void SAC_DISTMEM_PR_Barrier(void);

/******************************************
 * Functions that are implemented as
 * macros.
 *******************************************/

/** <!--********************************************************************-->
 *
 * @fn SAC_DISTMEM_TRUNCATED_START( var_NT, start)
 *
 *   @brief   Maximum of start and ShrayStart
 *
 *   @param  var_NT       array
 *   @param  start        index
 *
 *   @return              Maximum of start and ShrayStart
 *
 ******************************************************************************/

#define SAC_DISTMEM_TRUNCATED_START(var_NT, start)                            \
        SAC_MAX((intptr_t)ShrayStart(SAC_ADDRESS(var_NT)), start)

/** <!--********************************************************************-->
 *
 * @fn SAC_DISTMEM_TRUNCATED_END( var_NT, end)
 *
 *   @brief  Minimum of end and ShrayEnd
 *
 *   @param  var_NT       array
 *   @param  start        index
 *
 *   @return              Minimum of end and ShrayEnd
 *
 ******************************************************************************/

#define SAC_DISTMEM_TRUNCATED_END(var_NT, end)                                \
        SAC_MIN((intptr_t)ShrayEnd(SAC_ADDRESS(var_NT)), end)

/** <!--********************************************************************-->
 *
 * @fn SAC_DISTMEM_SHOULD_DIST( var_NT)
 *
 *   @brief   Calculates whether an array should be distributed, taking the
 *            execution mode into account.
 *
 *   @param  var_NT       the array in question
 *
 ******************************************************************************/

#define SAC_DISTMEM_SHOULD_DIST(var_NT)                                       \
    ((SAC_ND_A_DIM(var_NT) == 0) ? FALSE :                                    \
        SAC_DISTMEM_DET_DO_DISTR_ARR(SAC_ND_A_SIZE(var_NT),                   \
                SAC_ND_A_SHAPE(var_NT, 0)) &&                                 \
                (SAC_DISTMEM_exec_mode == SAC_DISTMEM_exec_mode_replicated))

/** <!--********************************************************************-->
 *
 * @fn SAC_DISTMEM_SWITCH_TO_DIST_EXEC()
 *
 *   @brief   Switches to distributed execution mode.
 *
 *            In distributed execution mode, each node is working on its share
 *            of a with-loop. If another with-loop is encountered while in
 *            distributed execution mode, it will not be distributed.
 *
 ******************************************************************************/

#define SAC_DISTMEM_SWITCH_TO_DIST_EXEC()                                     \
    SAC_TR_DISTMEM_PRINT ("Switching to distributed execution mode.");        \
    SAC_DISTMEM_CHECK_IS_SWITCH_TO_DIST_EXEC_ALLOWED ()                       \
    SAC_DISTMEM_exec_mode = SAC_DISTMEM_exec_mode_dist;

/** <!--********************************************************************-->
 *
 * @fn SAC_DISTMEM_SWITCH_TO_FOLD_EXEC()
 *
 *   @brief   Switches to distributed fold execution mode.
 *
 *            In distributed fold execution mode, each node is working on its
 *            share of a fold with-loop. If another with-loop is encountered
 *            while in distributed execution mode, it will not be distributed.
 *
 ******************************************************************************/

#define SAC_DISTMEM_SWITCH_TO_FOLD_EXEC()                                     \
    SAC_TR_DISTMEM_PRINT ("Switching to distributed fold execution mode.");   \
    SAC_DISTMEM_exec_mode = SAC_DISTMEM_exec_mode_fold;

/** <!--********************************************************************-->
 *
 * @fn SAC_DISTMEM_SWITCH_TO_REPLICATED_EXEC()
 *
 *   @brief   Switches to replicated execution mode.
 *
 *            In replicated execution mode, all nodes are executing the same
 *            program in parallel.
 *
 ******************************************************************************/

#define SAC_DISTMEM_SWITCH_TO_REPLICATED_EXEC()                               \
    SAC_TR_DISTMEM_PRINT ("Switching to replicated execution mode.");         \
    SAC_DISTMEM_CHECK_IS_SWITCH_TO_REPLICATED_EXEC_ALLOWED()                  \
    SAC_DISTMEM_exec_mode = SAC_DISTMEM_exec_mode_replicated;

/** <!--********************************************************************-->
 *
 * @fn void SAC_DISTMEM_SETUP(size_t min_elems_per_node, int trace_profile_rank);
 *
 *   @brief    Sets up the dsm system so that it is ready for use.
 *
 *   @param min_elems_per_node        minimum number of elements per node such
 *                                    that an array gets distributed
 *   @param trace_profile_rank        If not equal to
 *                                    SAC_DISTMEM_TRACE_PROFILE_RANK_ANY,
 *                                    only produce trace output for the
 *                                    provided rank.
 *
 ******************************************************************************/

#define SAC_DISTMEM_SETUP()                                                   \
    ShrayInit(NULL, NULL);                                                    \
    SAC_DISTMEM_trace_profile_rank = SAC_SET_DISTMEM_TRACE_PROFILE_NODE;      \
    SAC_DISTMEM_min_elems_per_node = SAC_SET_DISTMEM_MIN_ELEMS_PER_NODE;      \
    SAC_DISTMEM_rank = ShrayRank();                                           \
    SAC_DISTMEM_size = ShraySize();                                           \

/******************************************
 * Macros to hide trace settings
 *******************************************/

#if SAC_DO_TRACE_DISTMEM

#define SAC_DISTMEM_BARRIER()                                                 \
    SAC_DISTMEM_TR_Barrier ();                                                \

#define SAC_DISTMEM_DET_DO_DISTR_ARR(total_elems, dim0_size)                  \
    SAC_DISTMEM_TR_DetDoDistrArr (total_elems, dim0_size)

#define SAC_COMMLIB_BROADCAST(buffer, size, root)                             \
    SAC_COMMLIB_TR_Broadcast(buffer, size, root)

#define SAC_DISTMEM_ALLGATHER(value_type, value_NT, sub_results_NT)           \
    SAC_DISTMEM_TR_Allgather(SAC_ADDRESS(sub_results_NT),                     \
            SAC_ADDRESS(value_NT), sizeof(value_type))

#elif SAC_DO_PROFILE_DISTMEM

#define SAC_DISTMEM_BARRIER()                                                 \
    SAC_DISTMEM_PR_Barrier ();                                                \

#define SAC_DISTMEM_DET_DO_DISTR_ARR(total_elems, dim0_size)                  \
    SAC_DISTMEM_PR_DetDoDistrArr (total_elems, dim0_size)

#define SAC_COMMLIB_BROADCAST(buffer, size, root)                             \
    SAC_COMMLIB_PR_Broadcast(buffer, size, root)

#define SAC_DISTMEM_ALLGATHER(value_type, value_NT, sub_results_NT)           \
    SAC_DISTMEM_PR_Allgather(SAC_ADDRESS(sub_results_NT),                     \
            SAC_ADDRESS(value_NT), sizeof(value_type))

#else

#define SAC_DISTMEM_BARRIER() SAC_DISTMEM_Barrier ()

#define SAC_DISTMEM_DET_DO_DISTR_ARR(total_elems, dim0_size)                  \
    SAC_DISTMEM_DetDoDistrArr (total_elems, dim0_size)

#define SAC_COMMLIB_BROADCAST(buffer, size, root)                             \
    SAC_COMMLIB_Broadcast(buffer, size, root)

#define SAC_DISTMEM_ALLGATHER(value_type, value_NT, sub_results_NT)           \
    SAC_DISTMEM_Allgather(SAC_ADDRESS(sub_results_NT), SAC_ADDRESS(value_NT), \
            sizeof(value_type))

#endif

#include "runtime/distmem_h/side_effects.h"
#include "runtime/distmem_h/fold.h"
#include "runtime/distmem_h/prf.h"

#endif /* SAC_DO_DISTMEM */

#endif /* _SAC_DISTMEM_H */
