/******************************************
 * Macros for handling side-effect functions.
 ******************************************/

/* Creates variable names for the dimension and size of an AUD/AKD array or
 * hidden object created in side effects mode. This needs to be broadcasted to
 * the workers before they can allocate memory for the actual AUD/AKD array. */
#define _BC_DIMSIZE_VAR(value_NT) CAT12 (NT_NAME (value_NT), __bc_dimsize_var)
#define _BC_SIZE_VAR(value_NT) CAT12 (NT_NAME (value_NT), __bc_size_var)

/** <!--********************************************************************-->
 *
 * @fn SAC_DISTMEM_BROADCAST__WITH_DESC(value_NT, value_type)
 *
 *   @brief   This macro is for AKD and AUD arrays; the dimension, the size and
 *            the descriptor are broadcast as well. Broadcasts of distributed
 *            arrays are not supported but this is also not necessary since
 *            distributed arrays cannot be allocated in side effects execution
 *            mode.
 *
 *            Note: This macro is called in between of the function application
 *            and the update of the mirror. Since it is used for AKD and AUD
 *            arrays, (some of) the mirror variables have not yet been
 *            initialized. We need to initialize them before calling
 *            SAC_ND_ALLOC_BEGIN, however. After the descriptor has been
 *            allocated we copy the values we received from the master node.
 *
 *   @param value_type    C-type of the value
 *   @param value_NT      target variable
 *
 ******************************************************************************/

#define SAC_DISTMEM_BROADCAST__WITH_DESC(value_NT, value_type)                \
{                                                                             \
    SAC_TR_DISTMEM_PRINT ("Broadcast with descriptor "                        \
                          "of variable %s (type: %s).",                       \
                          NT_STR (value_NT), #value_type);                    \
    /* Communicate the size and dimension of the array to the workers so they
     * can allocate. */                                                       \
    size_t _BC_DIMSIZE_VAR (value_NT)[3];                                     \
    if (SAC_DISTMEM_rank == SAC_DISTMEM_RANK_SOURCE) {                        \
        _BC_DIMSIZE_VAR(value_NT)[0] =                                        \
            SAC_A_DIM_BEFORE_UPDATE_MIRROR(value_NT);                         \
        _BC_DIMSIZE_VAR(value_NT)[1] = SAC_ND_A_DESC_SIZE (value_NT);         \
        _BC_DIMSIZE_VAR(value_NT)[2] = SAC_ND_A_DESC_SHAPE (value_NT, 0);     \
        SAC_COMMLIB_BROADCAST(_BC_DIMSIZE_VAR (value_NT), 3 * sizeof(size_t), \
            SAC_DISTMEM_RANK_SOURCE);                                         \
        SAC_COMMLIB_BROADCAST(SAC_ND_A_DESC (value_NT),                       \
            BYTE_SIZE_OF_DESC (SAC_ND_A_DIM (value_NT)),                      \
            SAC_DISTMEM_RANK_SOURCE);                                         \
        SAC_COMMLIB_BROADCAST(SAC_ADDRESS(value_NT),                          \
            SAC_ND_A_DESC_SIZE(value_NT) * sizeof(value_type),                \
            SAC_DISTMEM_RANK_SOURCE);                                         \
    } else {                                                                  \
        SAC_COMMLIB_BROADCAST(_BC_DIMSIZE_VAR (value_NT), 3 * sizeof(size_t), \
            SAC_DISTMEM_RANK_SOURCE);                                         \
        SAC_UPDATE_A_MIRROR_DIM (value_NT, _BC_DIMSIZE_VAR (value_NT)[0]);    \
        SAC_ND_A_MIRROR_SIZE (value_NT) = _BC_DIMSIZE_VAR (value_NT)[1];      \
        SAC_ND_A_SHAPE (value_NT, 0) = _BC_DIMSIZE_VAR (value_NT)[2];         \
        SAC_ND_ALLOC_BEGIN (value_NT, 1, SAC_ND_A_DIM (value_NT), value_type) \
        SAC_COMMLIB_BROADCAST(SAC_ND_A_DESC (value_NT),                       \
            BYTE_SIZE_OF_DESC (SAC_ND_A_DIM (value_NT)),                      \
            SAC_DISTMEM_RANK_SOURCE);                                         \
        SAC_ND_ALLOC_END (value_NT, 1, SAC_ND_A_DIM (value_NT), value_type)   \
        SAC_COMMLIB_BROADCAST(SAC_ADDRESS(value_NT),                          \
            SAC_ND_A_SIZE (value_NT) * sizeof(value_type),                    \
            SAC_DISTMEM_RANK_SOURCE);                                         \
    }                                                                         \
}

/** <!--********************************************************************-->
 *
 * @fn SAC_DISTMEM_BROADCAST__AKS(value_NT, value_type)
 *
 *   @brief   This macro is for AKS arrays; the dimension, the size and the
 *            descriptor are not broadcast.
 *            Broadcasts of distributed arrays are not supported but this is
 *            also not necessary since distributed arrays cannot be allocated
 *            in side effects execution mode.
 *
 *            Before this macro is compiled in, we increment reference counters
 *            of distributed in arguments so that they are not freed within the
 *            function. The updated reference counters will be broadcast from
 *            the master to the workers after the function application.
 *            Then, all nodes will decrement the reference counters and if they
 *            become zero, free the variables. Distributed arrays are always
 *            AKS (for now).
 *
 *   @param value_type    C-type of the value
 *   @param value_NT      target variable
 *
 ******************************************************************************/
#define SAC_DISTMEM_BROADCAST__AKS(value_NT, value_type)                      \
    SAC_TR_DISTMEM_PRINT ("Broadcast of AKS variable %s (type: %s).",         \
            NT_STR (value_NT), #value_type);                                  \
    SAC_ND_ALLOC_BEGIN (value_NT, 1, SAC_ND_A_DIM (value_NT), value_type)     \
    SAC_ND_ALLOC_END (value_NT, 1, SAC_ND_A_DIM (value_NT), value_type)       \
    SAC_COMMLIB_BROADCAST(SAC_ADDRESS (value_NT), SAC_ND_A_SIZE (value_NT) *  \
            sizeof(value_type), SAC_DISTMEM_RANK_SOURCE);

/** <!--********************************************************************-->
 *
 * @fn SAC_DISTMEM_BROADCAST__SCL(value_NT, value_type)
 *
 *   @brief   This macro is for SCL; the dimension, the size and the descriptor
 *            are not broadcast.
 *
 *   @param value_type    C-type of the value
 *   @param value_NT      target variable
 *
 ******************************************************************************/

#define SAC_DISTMEM_BROADCAST__SCL(value_NT, value_type)                      \
    SAC_TR_DISTMEM_PRINT ("Scalar broadcast "                                 \
                          "of variable %s (type: %s)",                        \
                          NT_STR (value_NT), #value_type);                    \
    SAC_COMMLIB_BROADCAST(SAC_ADDRESS(value_NT),                              \
            sizeof(value_type), SAC_DISTMEM_RANK_SOURCE);                     \

/** <!--********************************************************************-->
 *
 * @fn SAC_DISTMEM_BROADCAST_RC( value_NT)
 *
 *   @brief   Broadcasts the reference counter of a distributed array, or does
 *            nothing if the array is not distributed.
 *
 *   @param value_NT      target variable
 *
 ******************************************************************************/

#define SAC_DISTMEM_BROADCAST_RC(value_NT)                                    \
        if (SAC_ND_A_DESC_IS_DIST (value_NT)) {                               \
            SAC_TR_DISTMEM_PRINT("Broadcast of reference count of %s",        \
                                 NT_STR(value_NT));                           \
            SAC_COMMLIB_BROADCAST(&SAC_ND_A_RC (value_NT), sizeof(intptr_t),  \
                    SAC_DISTMEM_RANK_SOURCE);                                 \
            SAC_ND_SET__RC (value_NT, SAC_ND_A_RC (value_NT));                \
        }
